#!/usr/bin/perl

$| = 1;

# ddrlogor.pl   ("DDR LOGfile logical .OR.")
#
# This is a Perl program which reads two GNU ddrescue 1.x "log files" (really,
# sector list files), and does a logical OR on them.  In other words, in the
# resulting output file, a sector is only shown as "unrecovered" if it was
# unrecovered in BOTH of the two input files.
#
# The purpose is to enable you to mark as "recovered" the unused/freespace
# files on partition.  Here's the strategy:
#
# 1) Run ddrescue to recover most of the data on a drive, into an image file.
# 2) Copy the image file onto a new drive.
# 3) Use ddr2nfi.pl to generate nficmds.bat, containing nfi commands for the
#    unrecovered sectors.
# 4) Attach the new drive to a Win2K or WinXP computer, and run nficmds.bat
#    on it, redirecting the output to a file on some other drive.  (You might
#    have to acknowledge a lot of NFI error popups, if the $Mft is damaged.)
# 5) Use 'ddrcruncher.pl -u' to generate unimportant.log, a ddrescue format
#    log file which shows all the "not in any file" bad sectors as recovered.
# 6) Use this program to merge unimportant.log into the real ddrescue log file.
#
# This can improve or speed up success with SpinRite (via ddr2sr.pl) and/or
# "freezer recovery."
#
#
# Note that if your log files contain both '-' and '/' intervals, then the
# operation is not quite symetrical:
#
#                  log_input1      log_input2      result
#                  ----------      ----------      ------
#  if sector is        +            anything          +
#  if sector is     anything           +              +
#  if sector is        -               -              -
#  if sector is        -               /              -
#  if sector is        /               -              /
#  if sector is        /               /              /


# BTW, here's a hint.  Open a console and do:
#   tail -s 3 -f /var/log/messages
# to watch the messages logged during a ddrescue attempt.


# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#


#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "5, 05-Oct-09";
# the TLIB Version Control version number and last-modified date


# options:
$verbose = 0;  # will be changed if '-v' is specified
$quick = 0;  # will be changed if '-q' is specified
$generalized_OR = 0;  # will be changed if '-o' is specified


$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}


@sv_ARGV = @ARGV;  # save raw parameter list, for display to the user

# display command-line parameters (just once)
$args_shown = 0;
sub showargs {
   if (!$args_shown) {
      print join( " ", "perl $0", @sv_ARGV ) . "\n";  # echo command-line arguments
      $args_shown = 1;
   }
}


# Handle input parameters:
while (($#ARGV >= 0) && ($ARGV[0] =~ /^-/)) {
   # option flag ("-")
   foreach $ch (split(//,$ARGV[0])) {
      if ('v' eq $ch) {
         $verbose++;
      } elsif ('q' eq $ch) {
         $quick = 1;
      } elsif ('o' eq $ch) {
         $generalized_OR = 1;
      } elsif ('-' ne $ch) {
         print STDERR "Unrecognized command-line option '-$ch' ignored.\n";
      }
   }
   shift @ARGV;
}
($log_output, $log_input1, $log_input2) = @ARGV;

if ($verbose) {
   &showargs;
   print "You are using Perl version $perlver\n";
}

if (!defined $log_input1) {
   die "ddrlogor.pl   (\"DDRescue LOGfile logical .OR.\")   v.$ver_ordinal_and_date\n" .
       "\n" .
       "Usage:\n" .
       "   perl -w $0 drive.log.new drive.log unimportant.log\n" .
       "\n" .
       "or, to specify options:\n" .
       "   perl -w $0 {options} drive.log.new drive.log unimportant.log\n" .
       "\n" .
       "{options} can be omitted or:\n" .
       "     -v   for verbose output (can be repeated for very verbose)\n" .
       "     -q   for quick mode (skip slow debug sanity checks)\n" .
       "     -o   for 'generalized .OR. mode' (just changes some output messages)\n" .
       "\n" .
       "e.g., for verbose output:\n" .
       "   perl -w $0 -v drive.log.new drive.log unimportant.log\n" .
       "\n" .
       "*** Note the odd parameter order (output file first)! ***\n" .
       "\n" .
       "See the comments in $0 for more detailed instructions.\n";
}


&showargs;

open( LOG_INPUT1, "<$log_input1") || die "ERROR: Could not read $log_input1, $!\n";
open( LOG_INPUT2, "<$log_input2") || die "ERROR: Could not read $log_input2, $!\n";



######################################################################
# Here's a sample of what a ddrescue log file looks like:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.2
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
######################################################################


# input is an integer represented as a string (probably hexadecimal preceeded by '0x'),
# result is the integer divided by 512.
sub divby512 {
   local($dividend) = @_;
   # print "dividend=$dividend\n";
   local($result) = 0;
   if ($dividend =~ /^0x/) {
      # print "tricky way!\n";
      # tricky divide by 512: drop the last 2 hexidecimal digits, and divide by 2 (which is more robust for very large integers)
      $result = substr($dividend,0,length($dividend)-2);
      $result = int(eval($result) / 2);  # gotta use eval() because it starts with 0x
   } else {
      # print "conventional way,\n";
      $result = int((0+$dividend) / 512);
   }
   # print "result=$result\n";
   return $result;
}


# Internal representation of the sector list is a bit tricky.
# The problem is that there are soooo many sectors.  A 300 GB
# disk drive has 600 million sectors.  Even at one character
# per sector, that's too big to represent in memory as a simple
# array or string.  So we have to represent it like it is in
# the sector list file, as a list of ranges.  But that means
# that we can't use a simple array or hash lookup to check the
# status of a particular sector; we'll have to call a function
# to do a search.



# First, read log_input1 (the old ddrescue sector list)
$numranges = 0;
$lastline = '';
$highest_from_inp1 = 0;
while (<LOG_INPUT1>) {
   if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
      $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
      # Parse a ddrescue log file line:
      @tmp = split( /\s+/, $_ );
      # for ($i=0; $i<=$#tmp; $i++) { print "dbg: tmp[$i]='" . $tmp[$i] . "'\n"; }
      ($pos_byt, $len_byt, $status) = @tmp;
      if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
         if ((1 == $#tmp) && (0 == $numranges) && ($len_byt =~ /^[\?\*\/\-\+]$/)) {
            # the new status line (introduced in ddrescue 1.6 or 1.7)
            $pos_sect = &divby512($pos_byt);
            $status = $len_byt;
            print "Status: ";
            if ($status eq '?') {
               print "copying data";
            } elsif ($status eq '*') {
               print "trimming bad areas";
            } elsif ($status eq '/') {
               print "splitting bad areas";
            } elsif ($status eq '-') {
               print "retrying bad blocks";
            } elsif ($status eq '+') {
               print "finished";
            } else {
               print "unrecognized status '$status'";
            }
            print " (status code = '$status'), at sector $pos_sect\n";
         } else {
            print "ERROR: unrecognized line in '$log_input1':\n$_\n";
            printf 'dbg: $#tmp=%d, $numranges=%d, $len_byt="%s"' . "\n", $#tmp, $numranges, $len_byt;
         }
         next;
      }
      $lastline = '';
      if ($len_byt eq '-1') {
         $lastline = $_;
         print "Note: length=-1 for status = '$status' in '$log_input1':\n$_\n";
         next;
      }

      # Now convert byte offsets and counts to LBA sector number & sector count:
      $pos_sect = &divby512($pos_byt);
      $len_sect = &divby512($len_byt);

      # print "dbg: Bad sect no. = $pos_sect, count = $len_sect, '$status'\n";
      # for ($i=0; $i<$len_sect; $i++) {
      #   $originally_bad_sectors{$pos_sect} = 1;
      #   $num_originally_bad_sectors++;
      #   $pos_sect++;
      # }

      $offsets[$numranges] = $pos_sect;
      $lengths[$numranges] = $len_sect;
      $stati[$numranges] = $status;  # '-' or '+' or '/'
      $numranges++;
      $highest_from_inp1 = $pos_sect + $len_sect;
   }
}
close(LOG_INPUT1);


$count_rescued1 = $count_failed1 = 0;
for ($i=0; $i<$numranges; $i++) {
   $status = $stati[$i];
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   if ('+' eq $status) {
      $count_rescued1 += $len;
   } elsif ('-' eq $status) {
      $count_failed1 += $len;
   }
}


# debug sanity-check code:
sub sanity_check {
  if ($quick) {
     return;
  }
  local($description) = shift;
  local($ofs_plus_len) = $lengths[0];
  local($i, $ofs, $len);
  for ($i=1; $i<$numranges; $i++) {
    $ofs = $offsets[$i];
    $len = $lengths[$i];
    if ($ofs < $ofs_plus_len) {
      print STDERR "ERROR: " . ($ofs_plus_len-$ofs) . "-sector overlap $description:\n  " .
                   ($i-1) . ": " . &displayrange($offsets[$i-1],$lengths[$i-1],$stati[$i-1]) . "\n  " .
		   $i . ": " . &displayrange($ofs,$len,$stati[$i]) . "\n";
      if ($verbose) {
         print "ERROR: " . ($ofs_plus_len-$ofs) . "-sector overlap $description:\n  " .
                   ($i-1) . ": " . &displayrange($offsets[$i-1],$lengths[$i-1],$stati[$i-1]) . "\n  " .
		   $i . ": " . &displayrange($ofs,$len,$stati[$i]) . "\n";
      }
      die;
    } elsif ($ofs > $ofs_plus_len) {
      print STDERR "ERROR: " . ($ofs-$ofs_plus_len) . "-sector gap $description:\n  " .
                   ($i-1) . ": " . &displayrange($offsets[$i-1],$lengths[$i-1],$stati[$i-1]) . "\n  " .
                   $i . ": " . &displayrange($ofs,$len,$stati[$i]) . "\n";
      if ($verbose) {
         print "ERROR: " . ($ofs-$ofs_plus_len) . "-sector gap $description:\n  " .
                   ($i-1) . ": " . &displayrange($offsets[$i-1],$lengths[$i-1],$stati[$i-1]) . "\n  " .
                   $i . ": " . &displayrange($ofs,$len,$stati[$i]) . "\n";
      }
    }
    $ofs_plus_len = $ofs + $len;
  }
}#sanity_check


&sanity_check( "in '$log_input1'" );


# Now read the second input file (probably unimportant.log, generated by nficruncher.pl)
# We ignore everything except '+' lines in the second input file.
$numranges2 = 0;
$lastline2 = '';
while (<LOG_INPUT2>) {
   if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
      $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
      # Parse a ddrescue log file line:
      @tmp = split( /\s+/, $_ );
      ($pos_byt, $len_byt, $status) = @tmp;
      if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
         print "ERROR: unrecognized line in '$log_input2':\n$_\n";
         next;
      }
      if ('+' ne $status) {
         # ignore all but '+' (recovered) regions
         $lastline2 = $_;
         next;
      }
      $lastline2 = '';

      # Now convert byte offsets and counts to LBA sector number & sector count:
      $pos_sect = &divby512($pos_byt);
      if ($len_byt eq '-1') {
         $len_sect = $highest_from_inp1;
      } else {
         $len_sect = &divby512($len_byt);
      }
      $offsets2[$numranges2] = $pos_sect;
      $lengths2[$numranges2] = $len_sect;
      $numranges2++;
   }
}
close(LOG_INPUT2);

# debug sanity-check code:
$ofs_plus_len = $lengths2[0];
for ($i=1; $i<$numranges2; $i++) {
   $ofs = $offsets2[$i];
   $len = $lengths2[$i];
   if ($ofs < $ofs_plus_len) {
      print STDERR "ERROR: " . ($ofs_plus_len-$ofs) . "-sector overlap in '$log_input2':\n  " .
                   $i-1 . ": " . &displayrange($offsets2[$i-1],$lengths2[$i-1],'+') . "\n  " .
                   $i . ": " . &displayrange($ofs,$len,'+') . "\n  " .
                   "(note: only '+' lines are counted)\n";
      if ($verbose) {
         print "ERROR: " . ($ofs_plus_len-$ofs) . "-sector overlap in '$log_input2':\n  " .
                   $i-1 . ": " . &displayrange($offsets2[$i-1],$lengths2[$i-1],'+') . "\n  " .
                   $i . ": " . &displayrange($ofs,$len,'+') . "\n  " .
                   "(note: only '+' lines are counted)\n";
      }
   }
   $ofs_plus_len = $ofs + $len;
}


# Format a ddrescue logfile line.  Input parameters are in sectors (512 bytes),
# but the result is displayed in bytes.
sub displayrange {
   local( $ofs, $len, $status ) = @_;
   local( $result ) = sprintf( "0x%06X00  0x%06X00  %s", (2*$ofs), (2*$len), $status );
   return $result;
}


# combine adjacent intervals with the same status (makes the log file smaller)
sub consolidate_adjacent {
   local( $i );
   if ($#offsets != ($numranges-1)) {
      print "ERR: numranges=%d   <>  (\$\#offsets+1)=%d\n", $numranges, 1+$#offsets;
   }
   for ( $i=1; $i<=$#offsets; ) {
      if (   ($offsets[$i] == ($offsets[$i-1]+$lengths[$i-1]))
          && ($stati[$i] eq $stati[$i-1]) ) {
         if ($verbose >= 2) {
             print "Consolidate:\n  " .
                   ($i-1) . ": " . &displayrange($offsets[$i-1],$lengths[$i-1],$stati[$i-1]) . "\n  " .
                   $i . ": " . &displayrange($offsets[$i],$lengths[$i],$stati[$i]) . "\n";
         }
         $lengths[$i-1] += $lengths[$i];
         splice( @offsets, $i, 1 );
         splice( @lengths, $i, 1 );
         splice( @stati, $i, 1 );
         $numranges--;
         if ($verbose >= 2) {
             print " yielding:\n  " .
                   ($i-1) . ": " . &displayrange($offsets[$i-1],$lengths[$i-1],$stati[$i-1]) . "\n";
         }
      } else {
         $i++;
      }
   }
}#consolidate_adjacent


if ($verbose) {
   print "Starting merge process.\n";
}

# Now we get to merge the two sector lists.
# Step through the (probably 'old') first file's '-' intervals, searching
# for overlaps with the (probably 'unimportant') 2nd file's '+' intervals
$j = 0;
for ($i=0; $i<$numranges; $i++) {
   # the only regions we care about are the '-' and '/' regions.
   if ('+' ne $stati[$i]) {
      $status = $stati[$i];
      $ofs = $offsets[$i];
      $len = $lengths[$i];
      if ($verbose) {
         print "dbg: check for overlap with file1.$i: " . &displayrange($ofs,$len,$status) . "\n";
      }
      # Was any part of this region marked with '+' in unimportant.log?
      while ($j<$numranges2) {
         $ofs2 = $offsets2[$j];
         $len2 = $lengths2[$j];
         if (($ofs2+$len2) <= $ofs) {
            # step to next interval in file2 ('unimportant')
            $j++;
            next;
         }
         if ($ofs2 >= ($ofs+$len)) {
            if ($verbose >= 2) {
               print "dbg: i=$i, no '+' region from file2 overlaps:\n" .
                     "  file1.$i: " . &displayrange($ofs,$len,$status) . "\n";
               if ($j > 0) {
                  print "  file2." . ($j-1) . ": " . &displayrange($offsets2[$j-1],$lengths2[$j-1],'+') . "\n";
               }
               print "  file2.$j: " . &displayrange($ofs2,$len2,'+') . "\n";
            }
            last;  # since the intervals are in order, we're done scanning file2 ('unimportant')
                   # for intervals that overlap this one from file1 ('old')
         }
         if ($verbose) {
            print "dbg: i=$i, j=$j, found an overlap.\n" .
                  "  file1.$i: " . &displayrange($ofs,$len,$status) . "\n" .
                  "  file2.$j: " . &displayrange($ofs2,$len2,'+') . "\n";
         }
         # They overlap!  Some of the sectors that were not successfully
         # recovered in the first input file, were unimportant (i.e., were
         # marked '+' in the second input file).
         # There are four cases:
         # 1) The unimportant interval invalidates the entire unrecovered region
         # 2) The unimportant interval invalidates the beginning of the unrecovered region
         # 3) The unimportant interval invalidates the end of the unrecovered region
         # 4) The unimportant interval invalidates the middle, but not the beginning or end
         if (($ofs2 <= $ofs) && (($ofs2+$len2) >= ($ofs+$len))) {
            # Case #1: the unimportant interval invalidates the entire unrecovered region
            $stati[$i] = $status = '+';
            if ($verbose) {
               print "  Entire 'old' (file1) range invalidated.\n";
               &sanity_check('after (case1)');
            }
            last;  # step to the next interval in file1
         } elsif ($ofs2 <= $ofs) {
            # Case #2: the unimportant interval invalidates the beginning of the unrecovered region
            # We can ignore the part that doesn't overlap:
            if ($ofs2 < $ofs) {
               $len2 -= ($ofs-$ofs2);
               $ofs2 = $ofs;
            }
            # split it
            splice( @offsets, $i+1, 0, $ofs2+$len2 );
            splice( @lengths, $i+1, 0, $len-$len2 );
            splice( @stati, $i+1, 0, $status );
            $stati[$i] = '+';
            $lengths[$i] = $len = $len2;
            $numranges++;
            if ($verbose) {
               print "  Beginning of old range invalidated.\n";
               &sanity_check('after (case2)');
            }
            last;  # step to next interval in file1 (actually the newly split-off 2nd half of this interval)
         } elsif (($ofs2+$len2) >= ($ofs+$len)) {
            # Case #3: the unimportant interval invalidates the end of the unrecovered region
            if (($ofs2+$len2) > ($ofs+$len)) {
               $len2 = $len+$ofs-$ofs2; # We can ignore the part that doesn't overlap:
            }
            # split it
            splice( @offsets, $i+1, 0, $ofs2 );
            splice( @lengths, $i+1, 0, $len2 );
            splice( @stati, $i+1, 0, '+' );
            $lengths[$i] = $len = ($len - $len2);
            $numranges++;
            if ($verbose) {
               print "  End of old range invalidated.\n  " .
                     $i . ': ' . &displayrange($ofs,$len,$status) . "\n  " .
                     ($i+1) . ': ' . &displayrange($offsets[$i+1],$lengths[$i+1],$stati[$i+1]) . "\n";
               &sanity_check('after (case3)');
            }
            last;  # step to next interval in file1 (actually the newly split-off 2nd half of this interval)
         } else {
            # Case #4 is the hardest, because we need to split the unrecovered region in two places
            splice( @offsets, $i+1, 0, $ofs2, $ofs2+$len2 );
            splice( @lengths, $i+1, 0, $len2, ($ofs+$len)-($ofs2+$len2) );
            splice( @stati, $i+1, 0, '+', $status );
            $lengths[$i] = $len = ($ofs2 - $ofs);
            $numranges += 2;
            if ($verbose) {
               print "  Middle of old range invalidated.  Split old range in 3 parts.\n";
               print "    i=$i, old len = $len\n";
               print "  Beg: " . &displayrange( $offsets[$i], $lengths[$i], $stati[$i] ) . "\n" .
                     "  Mid: " . &displayrange( $offsets[$i+1], $lengths[$i+1], '+' ) . "\n" .
                     "  End: " . &displayrange( $offsets[$i+2], $lengths[$i+2], $stati[$i+2] ) . "\n";
               &sanity_check('after (case4)');
            }
            last;  # step to next interval in file1 (actually the newly split-off 2nd & 3rd parts of this interval)
         }
      }
   }
}

if ($verbose) {
   print "Merge process complete.  Begin sanity-check.\n";
}
&sanity_check( 'in merged list' );
if ($verbose) {
   print "Sanity check complete.\n";
}


$count_rescued = $count_failed = 0;
for ($i=0; $i<$numranges; $i++) {
   $status = $stati[$i];
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   if ('+' eq $status) {
      $count_rescued += $len;
   } elsif ('-' eq $status) {
      $count_failed += $len;
   }
}
print "$count_rescued1 sectors were originally rescued (status '+' in '$log_input1').\n";
print "Plus, " . ($count_rescued-$count_rescued1) . " more were ";
if ($generalized_OR) {
   print "changed to status '+' by .OR.-ing '$log_input2'.\n";
} else {
   print "unimportant (free space) per '$log_input2'.\n";
}
print "So a total of $count_rescued sectors are now considered rescued (status '+' in '$log_output').\n";
print "Bad sector count reduced from $count_failed1 to $count_failed.\n";


# combine adjacent intervals which have the same status
if ($verbose) {
   print "Before consolidation, there were $numranges intervals.\n";
}
&consolidate_adjacent;
if ($verbose) {
   print "After consolidation, there are $numranges intervals.  Begin sanity-check.\n";
}
&sanity_check('in consolidated list');
if ($verbose) {
   print "Sanity check complete.  Now write the output file, '$log_output'...\n";
}


# Now spit out the list:
open( LOG_OUTPUT, ">$log_output" ) || die "ERROR: Could not create $log_output, $!\n";
binmode LOG_OUTPUT;  # write output in binary mode, to force LF-only line termination even under DOS/Windows
print LOG_OUTPUT "# Rescue Logfile. Created by GNU ddrescue version 1.2-burtonhack-ddrlogor.pl\n" .
      "#      pos        size  status\n";
$prev_ofs_plus_len = $prev_ofs = $prev_len = 0;
$prev_status = '?';
for ($i = 0; $i < $numranges; $i++) {
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   $status = $stati[$i];
   print LOG_OUTPUT &displayrange($ofs,$len,$status) . "\n";
   $prev_ofs_plus_len = $ofs + $len;
   $prev_ofs = $ofs;
   $prev_len = $len;
   $prev_status = $status;
}

if ($lastline ne '') {
   print LOG_OUTPUT "$lastline\n";  # last line with length of "-1"
}
print "'$log_output' created.\n";

__END__
